#!/usr/bin/env python
#-------------------------------------------------------------------
# tarfile.py
#
# Module for reading and writing .tar and tar.gz files.
#
# Needs at least Python version 2.2.
#
# Please consult the html documentation in this distribution
# for further details on how to use tarfile.
# 
#-------------------------------------------------------------------
# Copyright (C) 2002 Lars Gustbel <lars@gustaebel.de>
# All rights reserved.
#
# Permission  is  hereby granted,  free  of charge,  to  any person
# obtaining a  copy of  this software  and associated documentation
# files  (the  "Software"),  to   deal  in  the  Software   without
# restriction,  including  without limitation  the  rights to  use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies  of  the  Software,  and to  permit  persons  to  whom the
# Software  is  furnished  to  do  so,  subject  to  the  following
# conditions:
#
# The above copyright  notice and this  permission notice shall  be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS  IS", WITHOUT WARRANTY OF ANY  KIND,
# EXPRESS OR IMPLIED, INCLUDING  BUT NOT LIMITED TO  THE WARRANTIES
# OF  MERCHANTABILITY,  FITNESS   FOR  A  PARTICULAR   PURPOSE  AND
# NONINFRINGEMENT.  IN  NO  EVENT SHALL  THE  AUTHORS  OR COPYRIGHT
# HOLDERS  BE LIABLE  FOR ANY  CLAIM, DAMAGES  OR OTHER  LIABILITY,
# WHETHER  IN AN  ACTION OF  CONTRACT, TORT  OR OTHERWISE,  ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
#
"""Read from and write to tar format archives.
"""

__version__ = "$Revision: 1.29 $"
# $Source: /home/cvsroot/laas/tarfile/tarfile.py,v $

version = "0.4"
__author__ = "Lars Gustbel (lars@gustaebel.de)"
__date__ = "$Date: 2002/05/01 13:26:43 $"
__cvsid__ = "$Id: tarfile.py,v 1.29 2002/05/01 13:26:43 lars Exp $"
__credits__ = "Detlef Lannert for some early contributions"

#---------
# Imports
#---------
import sys
import os
import __builtin__
import shutil
import stat
import time
import cStringIO

try:
    import grp, pwd
except ImportError:
    grp = pwd = None

#------------------------
# TAR specific constants
#------------------------
BLOCKSIZE = 512                 # length of processing blocks
RECORDSIZE = BLOCKSIZE * 20     # length of records
MAGIC = "ustar"                 # magic tar string
VERSION = "00"                  # version number

LENGTH_NAME = 100               # maximal length of a filename
LENGTH_LINK = 100               # maximal length of a linkname

REGTYPE = "0"                   # regular file
AREGTYPE = "\0"                 # regular file
LNKTYPE = "1"                   # link (inside tarfile)
SYMTYPE = "2"                   # symbolic link
CHRTYPE = "3"                   # character special device
BLKTYPE = "4"                   # block special device
DIRTYPE = "5"                   # directory
FIFOTYPE = "6"                  # fifo special device
CONTTYPE = "7"                  # contiguous file

GNUTYPE_LONGNAME = "L"          # GNU tar extension for longnames
GNUTYPE_LONGLINK = "K"          # GNU tar extension for longlink
GNUTYPE_SPARSE = "S"            # GNU tar extension for sparse file

#-------------------------------
# tarfile specific constants
#-------------------------------
SUPPORTED_TYPES = (REGTYPE, AREGTYPE, LNKTYPE,  # file types that tarfile
                   SYMTYPE, DIRTYPE, FIFOTYPE,  # can cope with.
                   CONTTYPE, GNUTYPE_LONGNAME,
                   GNUTYPE_LONGLINK, GNUTYPE_SPARSE)

REGULAR_TYPES = (REGTYPE, AREGTYPE,             # file types that somehow
                 CONTTYPE, GNUTYPE_SPARSE)      # represent regular files

#-----------------------
# Some useful functions
#-----------------------
def nts(s):
    """Convert a null-terminated string buffer to a python string.
    """
    return s.split("\0", 1)[0]

def calc_chksum(buf):
    """Calculate the checksum for a member's header. It's a simple addition
    of all bytes, treating the chksum field as if filled with spaces.
    buf is a 512 byte long string buffer which holds the header.
    """
    chk = 256                           # chksum field is treated as blanks,
                                        # so the initial value is 8 * ord(" ")
    for c in buf[:148]: chk += ord(c)   # sum up all bytes before chksum
    for c in buf[156:]: chk += ord(c)   # sum up all bytes after chksum
    return chk

def copyfileobj(fo1, fo2, length=None):
    """Copy length bytes from fileobj fo1 to fileobj fo2.
    If length is None, copy the entire content.
    """
    if length == 0:
        return
    if length is None:
        shutil.copyfileobj(fo1, fo2)
        return
       
    BUFSIZE = 16 * 1024
    blocks, remainder = divmod(length, BUFSIZE)
    for b in range(blocks):
        buf = fo1.read(BUFSIZE)
        if not buf:
            raise IOError, "end of file reached"
        fo2.write(buf)

    buf = fo1.read(remainder)
    if not buf:
        raise IOError, "end of file reached"
    fo2.write(buf)
    return

if os.sep != "/":
    def normpath(path):
        return os.path.normpath(path).replace(os.sep, "/")
else:
    normpath = os.path.normpath

class TarError(Exception):
    """Internal exception"""
    pass

#--------------------
# exported functions
#--------------------
def open(name, mode="r", fileobj=None):
    """Open (uncompressed) tar archive name for reading, writing
    or appending.
    """
    return TarFile(name, mode, fileobj)

def gzopen(gzname, gzmode="r", compresslevel=9, fileobj=None):
    """Open gzip compressed tar archive name for reading or writing.
    Appending is not allowed.
    """
    if gzmode == "a":
        raise ValueError, "Appending to gzipped archive is not allowed"
    import gzip
    pre, ext = os.path.splitext(gzname)
    if ext == ".tgz":
        ext = ".tar"
    if ext == ".gz":
        ext = ""
    tarname = pre + ext
    mode = gzmode
    if "b" not in gzmode:
        gzmode += "b"
    if mode[0:1] == "w":
        if not fileobj:
            fileobj = __builtin__.file(gzname, gzmode)
        t = TarFile(tarname, mode, gzip.GzipFile(tarname, gzmode,
                                                 compresslevel, fileobj))
    else:
        t = TarFile(tarname, mode, gzip.open(gzname, gzmode, compresslevel))
    t._extfileobj = 0
    return t

def is_tarfile(name):
    """Return True if name points to a tar archive that we
    are able to handle, else return False.
    """

    # XXX replace 1 and 0 by True and False    
    buftoinfo = TarFile.__dict__["_buftoinfo"]
    try:
        buf = __builtin__.open(name, "rb").read(BLOCKSIZE)
        buftoinfo(None, buf)
        return 1
    except (ValueError, ImportError):
        pass
    try:
        import gzip
        buf = gzip.open(name, "rb").read(BLOCKSIZE)
        buftoinfo(None, buf)
        return 1
    except (IOError, ValueError, ImportError):
        pass
    return 0
    
#------------------
# Exported Classes
#------------------
class TarInfo:
    """Informational class which holds the details about an
    archive member given by a tar header block.
    TarInfo instances are returned by TarFile.getinfo() and
    TarFile.getmembers() and are usually created internally.
    If you want to create a TarInfo instance from the outside,
    you should use TarFile.gettarinfo() if the file already exists,
    or you can instanciate the class yourself.
    """

    def __init__(self, name=""):
        """Construct a TarInfo instance. name is the optional name
        of the member.
        """
        
        self.name = name             # member name (dirnames must end with '/')
        self.mode = 0100666          # file permissions
        self.uid = 0                 # user id
        self.gid = 0                 # group id
        self.size = 0                # file size
        self.mtime = 0               # modification time
        self.chksum = 0              # header checksum
        self.type = "0"              # member type
        self.linkname = ""           # link name
        self.uname = "user"          # user name
        self.gname = "group"         # group name
        self.devmajor = ""           #-+ not neccessary at this time
        self.devminor = ""           #-+ for use with CHRTYPE and BLKTYPE
        self.prefix = ""             # prefix, holding information
                                     # about sparse files
        
        self.offset = 0              # the tar header starts here
        self.offset_data = 0         # the optional file's data starts here
        
    def getheader(self):
        """Return a tar header block as a 512 byte string.
        """
        # The following code was contributed by Detlef Lannert.
        parts = []
        for value, fieldsize in (
                (self.name, 100),
                ("%07o" % self.mode, 8),
                ("%07o" % self.uid, 8),
                ("%07o" % self.gid, 8),
                ("%011o" % self.size, 12),
                ("%011o" % self.mtime, 12),
                ("        ", 8),
                (self.type, 1),
                (self.linkname, 100),
                (MAGIC, 6),
                (VERSION, 2),
                (self.uname, 32),
                (self.gname, 32),
                (self.devmajor, 8),
                (self.devminor, 8),
                (self.prefix, 155)
                ):
            l = len(value)
            parts.append(value + (fieldsize - l) * "\0")
        
        buf = "".join(parts)
        chksum = calc_chksum(buf)
        buf = buf[:148] + "%06o\0" % chksum + buf[155:]
        buf += (512 - len(buf)) * "\0"
        self.buf = buf
        return buf

    def isreg(self):
        return self.type in REGULAR_TYPES
    def isfile(self):
        return self.isreg()
    def isdir(self):
        return self.type == DIRTYPE
    def issym(self):
        return self.type == SYMTYPE
    def islnk(self):
        return self.type == LNKTYPE
    def ischr(self):
        return self.type == CHRTYPE
    def isblk(self):
        return self.type == BLKTYPE
    def isfifo(self):
        return self.type == FIFOTYPE
    def issparse(self):
        return self.type == GNUTYPE_SPARSE
# class TarInfo


class TarFile:
    """Class representing a TAR archive file on disk.
    """    
    debug = 0                   # may be set from 0 (no msgs) to 3 (all msgs)
    
    dereference = 0             # if true, add content of linked file to the
                                # tar file, else the link.
                                # XXX set to bool False

    def __init__(self, name=None, mode="r", fileobj=None):
        self.name = name
        
        if len(mode) > 1 or mode not in "raw":
            raise ValueError, "mode must be either 'r', 'a' or 'w', " \
                                "not '%s'" % mode
        self._mode = mode
        self.mode = {"r": "rb", "a": "r+b", "w": "wb"}[mode]
        
        if not fileobj:
            fileobj = __builtin__.file(self.name, self.mode)
            self._extfileobj = 0
        else:
            if self.name is None and hasattr(fileobj, "name"):
                self.name = fileobj.name
            if hasattr(fileobj, "mode"):
                self.mode = fileobj.mode
            self._extfileobj = 1
        self.fileobj = fileobj

        # Init datastructures
        self.members = []           # list of members as TarInfo instances
        self.membernames = []       # names of members 
        self.chunks = [0]           # chunk cache
        self._loaded = 0            # flag if all members have been read
        self.offset = 0l            # current position in the archive file
        self.inodes = {}            # dictionary caching the inodes of
                                    # archive members already added

        if self._mode == "a":
            self.fileobj.seek(0)
            self._load()

    def close(self):
        """Close the TarFile instance and do some cleanup.
        """
        if self.fileobj:
            if self._mode in "aw":
                # fill up the end with zero-blocks
                # (like option -b20 for tar does)
                blocks, remainder = divmod(self.offset, RECORDSIZE)
                if remainder > 0:
                    self.fileobj.write("\0" * (RECORDSIZE - remainder))

            if not self._extfileobj:
                self.fileobj.close()
            self.fileobj = None

    def next(self):
        """Return the next member from the archive.
        Return None if the end is reached.
        Can be used in a while statement, is used
        for Iteration (see __iter__()) and internally.
        """
        if not self.fileobj:
            raise ValueError, "I/O operation on closed file"
        if self._mode not in "ra":
            raise ValueError, "reading from a write-mode file"

        # Read the next block ignoring empty or "invalid" blocks.
        self.fileobj.seek(self.chunks[-1])
        while 1:
            buf = self.fileobj.read(BLOCKSIZE)
            if not buf:
                return None
            try:
                tarinfo = self._buftoinfo(buf)
            except ValueError:
                self.offset += BLOCKSIZE
                continue
            break

        # If the TarInfo instance contains a GNUTYPE longname or longlink
        # statement, we must parse it first.
        if tarinfo.type in (GNUTYPE_LONGLINK, GNUTYPE_LONGNAME):
            tarinfo = self._proc_gnulong(tarinfo, tarinfo.type)

        if tarinfo.issparse():
            # Sparse files need some care,
            # due to the possible extra headers.
            tarinfo.offset = self.offset
            self.offset += BLOCKSIZE
            origsize = self._proc_sparse(tarinfo)
            tarinfo.offset_data = self.offset
            blocks, remainder = divmod(tarinfo.size, BLOCKSIZE)
            if remainder:
                blocks += 1
            self.offset += blocks * BLOCKSIZE
            tarinfo.size = origsize
        else:
            tarinfo.offset = self.offset
            self.offset += BLOCKSIZE
            tarinfo.offset_data = self.offset
            if tarinfo.isreg():
                # Skip the following data blocks.
                blocks, remainder = divmod(tarinfo.size, BLOCKSIZE)
                if remainder:
                    blocks += 1
                self.offset += blocks * BLOCKSIZE

        self.members.append(tarinfo)
        self.membernames.append(tarinfo.name)
        self.chunks.append(self.offset)
        return tarinfo

    def getinfo(self, name):
        """Return a TarInfo instance for member name.
        """
        if name not in self.membernames and not self._loaded:
            self._load()
        if name not in self.membernames:
            raise KeyError, "filename `%s' not found in tar archive" % name
        return self._getmember(name)

    def getmembers(self):
        """Return a list of all members in the archive
        (as TarInfo instances).
        """
        if not self._loaded:    # if we want to obtain a list of
            self._load()        # all members, we first have to
                                # scan the whole archive.
        return self.members

    def getnames(self):
        """Return a list of names of all members in the
        archive.
        """
        if not self._loaded:
            self._load()
        return self.membernames

    def gettarinfo(self, name, arcname=None):
        """Create a TarInfo instance from an existing file.
        Optional arcname defines the name under which the file
        shall be stored in the archive.
        """
        # Building the name of the member in the archive.
        # Backward slashes are converted to forward slashes,
        # Absolute paths are turned to relative paths.
        if arcname is None:
            arcname = name
        arcname = normpath(arcname)
        drv, arcname = os.path.splitdrive(arcname)
        while arcname[0:1] == "/":
            arcname = arcname[1:]

        # Now, fill the TarInfo instance with
        # information specific for the file.
        tarinfo = TarInfo()

        # Use os.stat or os.lstat, depending on platform
        # and if symlinks shall be resolved.
        if hasattr(os, "lstat") and not self.dereference:
            osstat = os.lstat
        else:
            osstat = os.stat
            
        statres = osstat(name)
        linkname = ""

        stmd = statres.st_mode
        
        if stat.S_ISREG(stmd):
            inode = (statres.st_ino, statres.st_dev,
                     statres.st_mtime)
            if inode in self.inodes.keys() and not self.dereference:
                # Is it a hardlink to an already
                # archived file?
                type = LNKTYPE
                linkname = self.inodes[inode]
            else:
                # The inode is added only if its valid.
                # For win32 it is always 0.
                type = REGTYPE
                if inode[0]: self.inodes[inode] = arcname
                
        elif stat.S_ISDIR(stmd):
            type = DIRTYPE
            if arcname[-1:] != "/": arcname += "/"

        elif stat.S_ISFIFO(stmd):
            type = FIFOTYPE

        elif stat.S_ISLNK(stmd):
            type = SYMTYPE
            linkname = os.readlink(name)

        else:
            if self.debug > 1:
                print "'%s' has an unsupported type" % name
            return

        # Fill the TarInfo instance with all
        # information we can get.
        tarinfo.name = arcname
        tarinfo.mode = stmd
        tarinfo.uid = statres.st_uid
        tarinfo.gid = statres.st_gid
        tarinfo.size = statres.st_size
        tarinfo.mtime = statres.st_mtime
        tarinfo.type = type
        tarinfo.linkname = linkname
        if pwd:
            tarinfo.uname = pwd.getpwuid(tarinfo.uid)[0]
        if grp:
            tarinfo.gname = grp.getgrgid(tarinfo.gid)[0]
        return tarinfo        

    def list(self):
        """Print a formatted listing of the archive's
        contents to stdout.
        """
        print "%-46s %19s %12s" % ("File Name", "Modified    ", "Size")
        for tarinfo in self.getmembers():
            date = "%d-%02d-%02d %02d:%02d:%02d" \
                   % time.gmtime(tarinfo.mtime)[:6]
            name = tarinfo.name
            if len(name) > 46:
                name = name[:20] + "[...]" + name[-21:]
            size = str(tarinfo.size)
            if tarinfo.type not in REGULAR_TYPES:
                size = ""
            print "%-46s %s %12s" % (name, date, size)

    def add(self, name, arcname=None, recursive=1):
        """Add a file or a directory to the archive.
        Directory addition is recursive by default.
        """
        if not self.fileobj:
            raise ValueError, "I/O operation on closed file"
        if self._mode == "r":
            raise ValueError, "writing to a read-mode file"

        if arcname is None:
            arcname = name

        # Skip if somebody tries to archive the archive...
        if os.path.abspath(name) == os.path.abspath(self.name):
            if self.debug > 2:
                print name, "(skipped)"
            return

        # Special case: The user wants to add the current
        # working directory.
        if name == ".":
            if recursive:
                if arcname == ".":
                    arcname = ""
                for f in os.listdir("."):
                    self.add(f, os.path.join(arcname, f))
            return

        if self.debug > 1:
            print name

        # Create a TarInfo instance from the file.
        tarinfo = self.gettarinfo(name, arcname)

        # Append the tar header and data to the archive.
        if tarinfo.isreg():
            f = __builtin__.file(name, "rb")
            self.addfile(tarinfo, fileobj = f)
            f.close()

        if tarinfo.islnk() or tarinfo.issym() or tarinfo.isfifo():
            tarinfo.size = 0l
            self.addfile(tarinfo)

        if tarinfo.isdir():
            self.addfile(tarinfo)
            if recursive: 
                for f in os.listdir(name):
                    self.add(os.path.join(name, f), os.path.join(arcname, f))
    
    def addfile(self, tarinfo, fileobj=None):
        """Add the content of fileobj to the tarfile.
        The amount of bytes to read is determined by
        the size attribute in the tarinfo instance.
        """
        if not self.fileobj:
            raise ValueError, "I/O operation on closed file"
        if self._mode == "r":
            raise ValueError, "writing to a read-mode file"

        try:
            self.fileobj.seek(self.chunks[-1])
        except IOError:
            pass

        # Now we must check if the strings for filename
        # and linkname fit into the posix header.
        # (99 chars + "\0" for each)
        # If not, we must create GNU extension headers.
        # If both filename and linkname are too long,
        # the longlink is first to be written out.
        if len(tarinfo.linkname) >= LENGTH_LINK - 1:
            self._create_gnulong(tarinfo.linkname, GNUTYPE_LONGLINK)
            tarinfo.linkname = tarinfo.linkname[:LENGTH_LINK -1]
        if len(tarinfo.name) >= LENGTH_NAME - 1:
            self._create_gnulong(tarinfo.name, GNUTYPE_LONGNAME)
            tarinfo.name = tarinfo.name[:LENGTH_NAME - 1]
        
        header = tarinfo.getheader()
        self.fileobj.write(header)
        self.offset += BLOCKSIZE

        # If there's data to follow, append it.
        if fileobj is not None:
            copyfileobj(fileobj, self.fileobj, tarinfo.size)
            blocks, remainder = divmod(tarinfo.size, BLOCKSIZE)
            if remainder > 0:
                self.fileobj.write("\0" * (BLOCKSIZE - remainder))
                blocks += 1
            self.offset += blocks * BLOCKSIZE

        self.members.append(tarinfo)
        self.membernames.append(tarinfo.name)
        self.chunks.append(self.offset)

    def extractfile(self, member):
        """Extract member from the archive and return a file-like
        object. member may be a name or a TarInfo instance.
        """
        if not self.fileobj:
            raise ValueError, "I/O operation on closed file"
        if self._mode != "r":
            raise ValueError, "reading from a write-mode file"

        if isinstance(member, TarInfo):
            tarinfo = member
        else:
            tarinfo = self.getinfo(member)

        if tarinfo.isreg():
            return _FileObject(self, tarinfo)
        elif tarinfo.islnk() or tarinfo.issym():
            linkname = tarinfo.linkname
            return self.extractfile(self._getmember(linkname, tarinfo))
        else:
            return None
        
    def extract(self, member, path=""):
        """Extract member from the archive and write it to
        current working directory using its full pathname.
        If optional path is given, it is attached before the
        pathname.
        member may be a name or a TarInfo instance.
        """
        if not self.fileobj:
            raise ValueError, "I/O operation on closed file"
        if self._mode != "r":
            raise ValueError, "reading from a write-mode file"

        if isinstance(member, TarInfo):
            tarinfo = member
        else:
            tarinfo = self.getinfo(name)

        if self.debug > 1:
            print tarinfo.name,
        try:
            self._extract_member(tarinfo, os.path.join(path, tarinfo.name))
        except TarError, e:
            if self.debug > 1:
                print
                print "tarfile: '%s'" % e
            return
        if self.debug > 1:
            print

    def _extract_member(self, tarinfo, targetpath):
        """Extract the TarInfo instance tarinfo to a pysical
        file called targetpath.
        """
        # Fetch the TarInfo instance for the given name
        # and build the destination pathname, replacing
        # forward slashes to platform specific separators.
        if targetpath[-1:] == "/":
            targetpath = targetpath[:-1]
        targetpath = os.path.normpath(targetpath)

        if tarinfo.type not in SUPPORTED_TYPES:
            raise TarError, "Unsupported type `%s'" % tarinfo.type

        # Create all upper directories.
        try:
            os.makedirs(os.path.dirname(targetpath))
        except (IOError, OSError), e:
            pass

        if tarinfo.isdir():
            try:
                os.mkdir(os.path.basename(targetpath))
            except (IOError, OSError), e:
                raise TarError, "Could not create %s:\n%s" % (tarinfo.name, e)

        if tarinfo.type in REGULAR_TYPES:
            try:
                source = self.extractfile(tarinfo)
                target = __builtin__.file(targetpath, "wb")
                copyfileobj(source, target)
            except (IOError, OSError), e:
                raise TarError, e

        if tarinfo.isfifo():
            if hasattr(os, "mkfifo"):
                try:
                    os.mkfifo(targetpath)
                except (IOError, OSError), e:
                    raise TarError, e
            else:
                raise TarError, "fifo not supported by os"
            
        # The member is either a symbolic link or a hardlink
        # (a link inside the tarfile).
        # GNU tar behaviour is imitated. That means, if the link cannot
        # be created (because of an error or lacking support by filesystem),
        # we try to resolve the link. First we look for a valid member in the
        # tar archive, second we look in the filesystem.
        if tarinfo.issym() or tarinfo.islnk():
            linkpath = tarinfo.linkname
            linkpath = os.path.join(os.path.dirname(targetpath), linkpath)
            linkpath = os.path.normpath(linkpath)
            if self.debug > 1:
                print "->", linkpath,
            try:
                if tarinfo.issym():
                    os.symlink(linkpath, targetpath)
                else:
                    os.link(linkpath, targetpath)
            except (AttributeError, IOError, OSError), e:
                linkpath = os.path.join(os.path.dirname(tarinfo.name),
                                        tarinfo.linkname)
                linkpath = os.path.normpath(linkpath)
                linkpath = normpath(linkpath)
                try:
                    self._extract_member(self.getinfo(linkpath), targetpath)
                except (IOError, OSError, KeyError), e:
                    linkpath = os.path.normpath(linkpath)
                    try:
                        shutil.copy2(linkpath, targetpath)
                    except (IOError, OSError), e:
                        raise TarError, "Link could not be created"

        # Here we set the file's last modification time.
        # XXX If file is a directory, we do not set its mtime
        # because in most of the cases, it is reset to
        # current time when accessing files underneath.
        # Seems to make problems under win32 anyway.
        if not tarinfo.isdir() and not tarinfo.issym():
            try:
                os.utime(targetpath, (tarinfo.mtime, tarinfo.mtime))
            except (IOError, OSError), e:
                if self.debug > 2:
                    print "utime '%s'" % e

        # Here we set the ownership for the file.
        # We must be logged in as root. So, we
        # look for the owner on our system, and
        # if found, give it to him, else it's ours.
        if not tarinfo.issym() and pwd and os.getgid() == 0:
            try:
                g = grp.getgrnam(tarinfo.gname)[2]
            except KeyError:
                try:
                    g = grp.getgrgid(tarinfo.gid)[2]
                except KeyError:
                    g = os.getgid()
            try:
                u = pwd.getpwnam(tarinfo.uname)[2]
            except KeyError:
                try:
                    u = pwd.getpwuid(tarinfo.uid)
                except KeyError:
                    u = os.getuid()
            try:
                os.chown(targetpath, u, g)
            except (IOError, OSError), e:
                if self.debug > 2:
                    print "chown '%s'" % e

        # Here we set the file permissions.
        if not tarinfo.issym():
            try:
                os.chmod(targetpath, tarinfo.mode)
            except (IOError, OSError), e:
                if self.debug > 2:
                    print "chmod '%s'" % e

    def _getmember(self, name, tarinfo=None):
        """Find an archive member by name from bottom to top.
        If tarinfo is given, it is used as the starting point.
        """
        if tarinfo is None:
            end = len(self.members)
        else:
            end = self.members.index(tarinfo)
            
        for i in xrange(end - 1, -1, -1):
            if name == self.membernames[i]:
                return self.members[i]
    
    def _load(self):
        """Read through the entire archive file and look for readable
        members.
        """
        while 1:
            tarinfo = self.next()
            if tarinfo is None:
                break
        self._loaded = 1            
        return

    def __iter__(self):
        """Provide an iterator object.
        """
        if self._loaded:
            return iter(self.members)
        else:
            return TarIter(self)

    def _buftoinfo(self, buf):
        """Transform a 512 byte block to a TarInfo instance.
        """
        tarinfo = TarInfo()
        tarinfo.name = nts(buf[0:100])
        tarinfo.mode = int(buf[100:108], 8)
        tarinfo.uid = int(buf[108:116],8)
        tarinfo.gid = int(buf[116:124],8)
        tarinfo.size = long(buf[124:136], 8)
        tarinfo.mtime = long(buf[136:148], 8)
        tarinfo.chksum = int(buf[148:156], 8)
        tarinfo.type = buf[156:157]
        tarinfo.linkname = nts(buf[157:257])
        tarinfo.uname = nts(buf[265:297])
        tarinfo.gname = nts(buf[297:329])
        tarinfo.devmajor = buf[329:337]
        tarinfo.devminor = buf[337:345]
        tarinfo.prefix = buf[345:500]
        if self.debug > 1 and tarinfo.chksum != calc_chksum(buf):
            print "tarfile: 'Bad Checksum'"
        return tarinfo

    def _proc_gnulong(self, tarinfo, type):
        """Evaluate the two blocks that hold a GNU longname
        or longlink member.
        """
        name = None
        linkname = None
        buf = self.fileobj.read(BLOCKSIZE)
        if not buf:
            return None
        self.offset += BLOCKSIZE
        if type == GNUTYPE_LONGNAME:
            name = nts(buf)
        if type == GNUTYPE_LONGLINK:
            linkname = nts(buf)

        buf = self.fileobj.read(BLOCKSIZE)
        if not buf:
            return None
        tarinfo = self._buftoinfo(buf)
        if name is not None:
            tarinfo.name = name
        if linkname is not None:
            tarinfo.linkname = linkname
        self.offset += BLOCKSIZE
        return tarinfo

    def _create_gnulong(self, name, type):
        """Insert a GNU longname/longlink member into the archive.
        It consists of a common tar header, with the length
        of the longname as size, followed by a data block,
        which contains the longname as a null terminated string.
        """
        tarinfo = TarInfo()
        tarinfo.name = "././@LongLink"
        tarinfo.type = type
        tarinfo.mode = 0
        tarinfo.size = len(name)

        self.fileobj.write(tarinfo.getheader())
        self.fileobj.write(name + "\0" * (512 - len(name)))
        self.offset += BLOCKSIZE * 2
    
    def _proc_sparse(self, tarinfo):
        """Analyze a GNU sparse header plus extra headers.
        """
        buf = tarinfo.getheader()
        sp = _ringbuffer()
        pos = 386
        lastpos = 0l
        realpos = 0l
        try:
            # There are 4 possible sparse structs in the
            # first header.
            for i in range(4):
                offset = int(buf[pos:pos + 12], 8)
                numbytes = int(buf[pos + 12:pos + 24], 8)
                if offset > lastpos:
                    sp.append(_hole(lastpos, offset - lastpos))
                sp.append(_data(offset, numbytes, realpos))
                realpos += numbytes
                lastpos = offset + numbytes
                pos += 24

            isextended = ord(buf[482])
            origsize = int(buf[483:495], 8)

            # If the isextended flag is given,
            # there are extra headers to process.
            while isextended == 1:
                buf = self.fileobj.read(BLOCKSIZE)
                self.offset += BLOCKSIZE
                pos = 0
                for i in range(21):
                    offset = int(buf[pos:pos + 12], 8)
                    numbytes = int(buf[pos + 12:pos + 24], 8)
                    if offset > lastpos:
                        sp.append(_hole(lastpos, offset - lastpos))
                    sp.append(_data(offset, numbytes, realpos))
                    realpos += numbytes
                    lastpos = offset + numbytes
                    pos += 24
                isextended = ord(buf[504])
        except ValueError:
            pass
        if lastpos < origsize:
            sp.append(_hole(lastpos, origsize - lastpos))

        tarinfo.sparse = sp
        return origsize
# class TarFile

# Helper classes for sparse file support
class _section:
    """Base class for _data and _hole.
    """
    def __init__(self, offset, size):
        self.offset = offset
        self.size = size
    def __contains__(self, offset):
        return self.offset <= offset < self.offset + self.size
class _data(_section):
    """Represent a data section in a sparse file.
    """
    def __init__(self, offset, size, realpos):
        _section.__init__(self, offset, size)
        self.realpos = realpos
class _hole(_section):
    """Represent a hole section in a sparse file.
    """
    pass
class _ringbuffer(list):
    """Ringbuffer class which increases performance
    over a regular list.
    """
    def __init__(self):
        self.idx = 0
    def find(self, offset):
        idx = self.idx
        while 1:
            item = self[idx]
            if offset in item:
                break
            idx += 1
            if idx == len(self):
                idx = 0
            if idx == self.idx:
                # End of File
                return None
        self.idx = idx
        return item

class _FileObject:
    """File-like object for reading an archive member,
    is returned by TarFile.extractfile().
    Support for sparse files included.
    """

    def __init__(self, tarfile, tarinfo):
        self.fileobj = tarfile.fileobj
        self.name = tarinfo.name
        self.mode = "r"
        self.closed = 0
        self.offset = tarinfo.offset_data
        self.size = tarinfo.size
        self.pos = 0l
        self.linebuffer = ""
        if tarinfo.issparse():
            self.sparse = tarinfo.sparse
            self.read = self._readsparse
        else:
            self.read = self._readnormal

    def readline(self, size=-1):
        """Read a line with approx. size.
        If size is negative, read a whole line.
        readline() and read() must not be mixed up (!).
        """
        if size < 0:
            size = sys.maxint

        nl = self.linebuffer.find("\n")
        if nl >= 0:
            nl = min(nl, size)
        else:
            size -= len(self.linebuffer)
            while nl < 0:
                buf = self.read(min(size, 100))
                if not buf:
                    break
                self.linebuffer += buf
                size -= len(buf)
                if size <= 0:
                    break
                nl = self.linebuffer.find("\n")
            if nl == -1:
                s = self.linebuffer
                self.linebuffer = ""
                return s
        buf = self.linebuffer[:nl]
        self.linebuffer = self.linebuffer[nl + 1:]
        while buf[-1:] == "\r":
            buf = buf[:-1]
        return buf + "\n"

    def readlines(self):
        """Return a list with all (following) lines.
        """
        result = []
        while 1:
            line = self.readline()
            if not line: break
            result.append(line)
        return result
                
    def _readnormal(self, size=None):
        """Read operation for regular files.
        """
        if self.closed:
            raise ValueError, "I/O operation on closed file"
        self.fileobj.seek(self.offset + self.pos)
        bytesleft = self.size - self.pos
        if size is None:
            bytestoread = bytesleft
        else:
            bytestoread = min(size, bytesleft)
        self.pos += bytestoread
        return self.fileobj.read(bytestoread)

    def _readsparse(self, size=None):
        """Read operation for sparse files.
        """
        if self.closed:
            raise ValueError, "I/O operation on closed file"

        if size is None:
            size = self.size - self.pos

        data = ""
        while size > 0:
            buf = self._readsparsesection(size)
            if not buf:
                break
            size -= len(buf)
            data += buf
        return data

    def _readsparsesection(self, size):
        """Read a single section of a sparse file.
        """
        section = self.sparse.find(self.pos)
        
        if section is None:
            return ""
        
        toread = min(size, section.offset + section.size - self.pos)
        if isinstance(section, _data):
            realpos = section.realpos + self.pos - section.offset
            self.pos += toread
            self.fileobj.seek(self.offset + realpos)
            return self.fileobj.read(toread)
        else:
            self.pos += toread
            return "\0" * toread

    def tell(self):
        """Return the current file position.
        """
        return self.pos

    def seek(self, pos, whence=0):
        """Seek to a position in the file.
        """
        self.linebuffer = ""
        if whence == 0:
            self.pos = min(max(pos, 0), self.size)
        if whence == 1:
            if pos < 0:
                self.pos = max(self.pos + pos, 0)
            else:
                self.pos = min(self.pos + pos, self.size)
        if whence == 2:
            self.pos = max(min(self.size + pos, self.size), 0)

    def close(self):
        """Close the file object.
        """
        self.closed = 1
#class _FileObject

#---------------------------------------------
# zipfile compatible TarFile class
#
# for details consult zipfile's documentation
#---------------------------------------------
TAR_PLAIN = 0           # zipfile.ZIP_STORED
TAR_GZIPPED = 8         # zipfile.ZIP_DEFLATED
class TarFileCompat:
    """TarFile class compatible with standard module zipfile's
    ZipFile class.
    """
    def __init__(self, file, mode="r", compression=TAR_PLAIN):
        if compression == TAR_PLAIN:
            self.tarfile = open(file, mode)
        elif compression == TAR_GZIPPED:
            self.tarfile = gzopen(file, mode)
        else:
            raise ValueError, "unknown compression constant"
        if mode[0:1] == "r":
            import time
            members = self.tarfile.getmembers()
            for i in range(len(members)):
                m = members[i]
                m.filename = m.name
                m.file_size = m.size
                m.date_time = time.gmtime(m.mtime)[:6]
    def namelist(self):
        return map(lambda m: m.name, self.infolist())
    def infolist(self):
        return filter(lambda m: m.type in REGULAR_TYPES,
                      self.tarfile.getmembers())
    def printdir(self):
        self.tarfile.list()
    def testzip(self):
        return
    def getinfo(self, name):
        return self.tarfile.getinfo(name)
    def read(self, name):
        return self.tarfile.extractfile(self.tarfile.getinfo(name)).read()
    def write(self, filename, arcname=None, compress_type=None):
        self.tarfile.add(filename, arcname)
    def writestr(self, zinfo, bytes):
        import calendar
        zinfo.name = zinfo.filename
        zinfo.size = zinfo.file_size
        zinfo.mtime = calendar.timegm(zinfo.date_time)
        self.tarfile.addfile(zinfo, cStringIO.StringIO(bytes))
    def close(self):
        self.tarfile.close()
#class TarFileCompat    

class TarIter:
    """Iterator Class.
    
    for tarinfo in TarFile(...):
        suite...
    """
    
    def __init__(self, tarfile):
        """Construct a TarIter instance.
        """
        self.tarfile = tarfile
    def __iter__(self):
        """Return iterator object.
        """
        return self
    def next(self):
        """Return the next item using TarFile's next() method.
        When all members have been read, set TarFile as _loaded.
        """
        tarinfo = self.tarfile.next()
        if not tarinfo:
            self.tarfile._loaded = 1
            raise StopIteration
        return tarinfo
# class TarIter

if __name__ == "__main__":
    # a "light-weight" implementation of GNUtar ;-)
    usage = """
Usage: %s [options] [files]

-h      display this help message
-c      create a tarfile
-r      append to an existing archive
-x      extract archive
-t      list archive contents
-f FILENAME
        use archive FILENAME, else STDOUT (-c)
-z      filter archive through gzip
-C DIRNAME
        with opt -x:     extract to directory DIRNAME
        with opt -c, -r: put files to archive under DIRNAME
-q      quiet

wildcards *, ?, [seq], [!seq] are accepted.
    """ % sys.argv[0]
    
    import getopt, glob
    try:
        opts, args = getopt.getopt(sys.argv[1:], "htcrzxf:C:q")
    except getopt.GetoptError, e:
        print
        print "ERROR:", e
        print usage
        sys.exit(0)
    
    file = None
    mode = None
    dir = ""
    comp = 0
    debug = 3
    for o, a in opts:
        if o == "-t": mode = "l"        # list archive
        if o == "-c": mode = "w"        # write to archive
        if o == "-r": mode = "a"        # append to archive
        if o == "-x": mode = "r"        # extract from archive
        if o == "-f": file = a          # specify filename else use stdout
        if o == "-C": dir = a           # change to dir
        if o == "-z": comp = 1          # filter through gzip
        if o == "-q": debug = 0         # quiet mode
        if o == "-h":                   # help message
            print usage
            sys.exit(0)

    if not mode:
        print usage
        sys.exit(0)

    if comp:
        func = gzopen
    else:
        func = open
    
    if not file:
        if mode != "w":
            print usage
            sys.exit(0)
        debug = 0
        # If under Win32, set stdout to binary.
        try:
            import msvcrt
            msvcrt.setmode(1, os.O_BINARY)
        except: pass
        tarfile = func("sys.stdout.tar", mode, sys.stdout)
    else:
        if mode == "l":
            tarfile = func(file, "r")
        else:
            tarfile = func(file, mode)

    tarfile.debug = debug

    if mode == "r":
        for tarinfo in tarfile:
            tarfile.extract(tarinfo, dir)
    elif mode == "l":
        tarfile.list()
    else:
        for arg in args:
            files = glob.glob(arg)
            for f in files:
                tarfile.add(f, dir)
    tarfile.close()
